﻿using Newtonsoft.Json;
using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Drawing;
using System.IO;
using System.Linq;
using System.Net;
using System.Net.NetworkInformation;
using System.Net.Sockets;
using System.Text;
using System.Threading;
using System.Windows.Forms;
using System.Text.RegularExpressions;
using WindowsUtilities;

namespace CTNet
{
    public enum verifyFile { fileOk, exception, nullEmptyPath, nullEmptyFileName, nullWorkFolder, notDocFile, noZipFile };
    public enum boxSelMsgTypes { boxSets, delBoxSet, delAllBoxSets, updateSel, newCHSet, newCHSelIndex, newCBCSelIndex,  setCBCState };

    // Things we tell the form about:
    public enum ctEvent { connected, connectFail, boxSel, boxSelAck, boxSelOff, shutdown, docLockOrNoCheckboxes, disconnectCH };

    // A custom EventArgs class to pass ctClient info to main form.
    public class CtClientArgs : EventArgs
    {
        public ctEvent eventType { get; set; }
        public string eventMsg { get; set; }
    }

    public class CTClient
    {
        // Create the calc tool client event itself:
        public event EventHandler<CtClientArgs> CtClientEvent;

        /*
         * We set this when asking for checkbox selections from
         * CalcHelper. If we receive a CHBoxSelOff message while
         * waiting we know that CalcHelper is telling us that it's
         * associated document has no checkboxes that it can possibly
         * select. We would also receive a CHBoxSelOff message if
         * CalcHelper is locked. This prevents a timeout while waiting
         * for the box selection request acknowledgment from
         * CalcHelper. Had to do it this way as a modification to
         * the Calc Tool Server is not possible.
         */
        public bool checkboxAckWait = false;

        public bool boxSelFromCH = false;

        private static bool bExiting = false;
        private static bool bConnected = false;

        /*
         * Saw a problem while testing trying to connect to
         * Calc Tool Server when I knew it was not running. Got an exception in
         * the ConnectCallback() routine which caused StartClient() to hang on
         * the connectDone.WaitOne() call since there was no connectDone.Set()
         * called from ConnectCallback(). So, I am calling connectDone.Set()
         * in a finally block in ConnectCallback() along with setting this
         * exception flag and will check it in StartClient() to make sure we
         * have a 'real' connectDone.Set() signal.
         */
        public bool connException = false;

        private static int port = 0;

        private static Socket client = null;

        public VSCFileInfo vscFileInfo = null;

        // This is the arg[0] from Visual Studio Code launch.
        public string vscFile = null;

        public static string vscFileNameNoExt = null;

        // Location of Calc Tool Server port number file so client tool can connect.
        private static string portFile = null;

        private static IntPtr _hWnd = IntPtr.Zero;

        private static CTMessage gMsg = new CTMessage();

        // Thread signal. Probably will not use some of these..... 
        private static ManualResetEvent allDone = new ManualResetEvent(false);
        private static ManualResetEvent msgDone = new ManualResetEvent(false);

        private static ManualResetEvent sendMsgDone = new ManualResetEvent(false);
        private static ManualResetEvent receiveMsgDone = new ManualResetEvent(false);

        private static ManualResetEvent sendDone = new ManualResetEvent(false);
        private static ManualResetEvent connectDone = new ManualResetEvent(false);
        private static ManualResetEvent connectAckDone = new ManualResetEvent(false);

        public CTClient(VSCFileInfo vscodeFileInfo)
        {
            vscFileInfo = vscodeFileInfo;

            SetFileInfo(true);
        }

        private void SetFileInfo(bool doPortFile)
        {
            /*
             * I think this is called 'normalizing' the file path,
             * so c:\blah\blah and c:/blah/blah are comparable
             * to each other. Needed when we get a file 
             * open/close/save message from Vscode.
             */
            vscFile = Path.GetFullPath(vscFileInfo.launchFileTmp).ToLower();

            vscFileNameNoExt = vscFileInfo.fileNameNoExtTmp;

            if (doPortFile == true)
            {
                portFile = vscFileInfo.portFileTmp;
            }
        }

        // Trigger the calc tool client event:
        protected virtual void OnCtClientEvent(CtClientArgs e)
        {
            if (CtClientEvent != null)
            {
                CtClientEvent(this, e);
            }
        }

        public bool IsConnected()
        {
            bool result = false;

            if (client != null && client.Connected == true)
            {
                result = true;
            }

            return result;
        }

        public void StopClient()
        {
            bExiting = true;
            receiveMsgDone.Set();

            if (client != null && client.Connected == true)
            {
                try
                {
                    client.Shutdown(SocketShutdown.Both);
                    client.Close();
                }
                catch
                {

                }
            }
        }

        public void StartClient()
        {
            CtClientArgs ctClientArgs = new CtClientArgs();

            // Connect to Visual Studio Code server.
            if (ReadPortNumber() == false)
            {
                MessageBox.Show("  Could not read .calctool port number file.\n\n  This is required for commuication with\n  the Calc Helper tool.\n\n  Aborting Calc Tool Server Connection Attempt.", "Calc Tool Server Connection Error:", MessageBoxButtons.OK, MessageBoxIcon.Error);

                ctClientArgs.eventType = ctEvent.connectFail;
            }
            else
            {
                try
                {
                    // Establish the remote endpoint for the socket.  
                    IPAddress ipAddress = IPAddress.Loopback;
                    IPEndPoint remoteEP = new IPEndPoint(ipAddress, port);

                    // Create a TCP/IP socket.  
                    client = new Socket(ipAddress.AddressFamily, SocketType.Stream, ProtocolType.Tcp);

                    // Connect to the remote endpoint.  
                    client.BeginConnect(remoteEP, new AsyncCallback(ConnectCallback), client);

                    connectDone.WaitOne();

                    // Debug:
                    // Added this if (..) { .. } else { ... } to see if we can abort and close
                    // this application if ConnectCallback() has an exception.
                    if (connException)
                    {
                        //MessageBox.Show("  We have a connException!", "Check Box Control-StartClient():", MessageBoxButtons.OK, MessageBoxIcon.Information);
                        ctClientArgs.eventType = ctEvent.connectFail;
                    }
                    else
                    {
                        // Send connect message to Calc Tool Server.
                        SendConnectMsg();

                        if (sendMsgDone.WaitOne(10000) == true)
                        {
                            // Now we expect the Connection Acknowledgment.
                            //MessageBox.Show("  Waiting for connection acknowledgment from Calc Tool Server.", "Check Box Control-StartClient():", MessageBoxButtons.OK, MessageBoxIcon.Information);

                            receiveMsgDone.Reset();

                            ReceiveAsync(client);

                            if (receiveMsgDone.WaitOne(30000) == true)
                            {
                                // Signal main form that we are connected to Calc Tool Server.
                                ctClientArgs.eventType = ctEvent.connected;

                                //MessageBox.Show("  Calc Tool Server has acknowledged our connection!.", "Check Box Control-StartClient():", MessageBoxButtons.OK, MessageBoxIcon.Information);
                            }
                            else
                            {
                                MessageBox.Show("  Timed out connecting to Calc Tool Server\n  while executing receiveMsgDone.WaitOne(30000).\n\n  This is required for commuication with\n  the Calc Helper tool.\n\n  Aborting Calc Tool Server Connection Attempt.", "Calc Tool Server Connection Error:", MessageBoxButtons.OK, MessageBoxIcon.Error);

                                ctClientArgs.eventType = ctEvent.connectFail;
                            }
                        }
                        else
                        {
                            MessageBox.Show("  Timed out connecting to Calc Tool Server\n  while executing sendMsgDone.WaitOne(10000).\n\n  This is required for commuication with\n  the Calc Helper tool.\n\n  Aborting Calc Tool Server Connection Attempt.", "Calc Tool Server Connection Error:", MessageBoxButtons.OK, MessageBoxIcon.Error);

                            ctClientArgs.eventType = ctEvent.connectFail;
                        }
                    }
                }
                catch (Exception ex)
                {
                    MessageBox.Show(ex.ToString(), "Calc Tool Server Connection Error:", MessageBoxButtons.OK, MessageBoxIcon.Error);
                    ctClientArgs.eventType = ctEvent.connectFail;
                }
            }

            OnCtClientEvent(ctClientArgs);
        }

        // From Joe's SBClient:
        public void SendConnectMsg()
        {
            CTMessage msg = new CTMessage();
            msg.Command = "Connect";

            SendAsync(msg);
        }

        /*
         * This is how Checkbox Control sends the request for
         * checkbox sets to Calc Helper. We reuse the "CBCBoxSel"
         * command and set the filename to the request text.
         * 
         * A little clunky but it works - unable to modify the
         * Calc Tool Server commands.
         */
        public void BoxSetsRequest()
        {
            if (client != null && client.Connected == true)
            {
                CTMessage msg = new CTMessage();

                msg.Command = "CBCBoxSel";
                msg.FileName = "ChkBoxSetsReq";
                SendAsync(msg);
            }
        }

        // Ask Calc Helper for check box selections.
        public void BoxSelRequest()
        {
            //string debug = "";
            bool showError = false;

            if (client != null && client.Connected == true)
            {
                CTMessage msg = new CTMessage();

                checkboxAckWait = true;

                msg.Command = "CBCBoxSel";

                //MessageBox.Show(string.Format("Asking for boxes, vscFile: {0}", vscFile), "CHBoxSelRequest():", MessageBoxButtons.OK, MessageBoxIcon.Information);

                msg.FileName = vscFile;

                sendMsgDone.Reset();

                SendAsync(msg);

                if (sendMsgDone.WaitOne(5000) == true)
                {
                    // DEBUG:
                    //MessageBox.Show("sendMsgDone.WaitOne(5000) - No timeout!", "CHBoxSelRequest():", MessageBoxButtons.OK, MessageBoxIcon.Information);

                    // Expecting an Ack from Calc Helper:
                    receiveMsgDone.Reset();

                    //if (receiveMsgDone.WaitOne(5000) == true)
                    if (!receiveMsgDone.WaitOne(5000))
                    {
                        //MessageBox.Show("receiveMsgDone.WaitOne(5000) - No timeout!", "CHBoxSelRequest():", MessageBoxButtons.OK, MessageBoxIcon.Information);
                        /*
                         * Signal main form that Calc Helper has acknowledged our
                         * request for check box selections.
                         */
                        //CtClientArgs ctClientArgs = new CtClientArgs();

                        //ctClientArgs.eventType = ctEvent.boxSelAck;
                        //OnCtClientEvent(ctClientArgs);
                        //}
                        //else
                        //{
                        showError = true;
                        //debug = "receiveMsgDone timeout";
                    }
                }
                else
                {
                    showError = true;
                    //debug = "sendMsgDone timeout";
                }

                if (showError)
                {
                    MessageBox.Show("  The message asking Calc Helper for check box selections\n  has timed out.\n\n  Make sure Calc Helper is running and the Visual Studio\n  Code icon is colored GREEN.", "Box Selection Request Timeout Error:", MessageBoxButtons.OK, MessageBoxIcon.Error);

                    // Debug.
                    //MessageBox.Show(string.Format("Timeout manual reset: {0}", debug), "Message Timeout Error:", MessageBoxButtons.OK, MessageBoxIcon.Information);
                }
            }
        }

        /*
         * Not using this, keeping it anyways. As it turned out
         * there is no way Checkbox Control can change it's
         * editor launched file. That CAN happen with Calc
         * Helper and is handled by Calc Helper going into a
         * lockdown state which breaks the link to Checkbox
         * Control. The user will be able to ask once more for
         * checkbox selections once Calc Helper reconnects to
         * the editor launch file - the exact same file that
         * Checkbox Control is 'imprinted' on when launched from
         * the editor.
         * 
         * Send Calc Helper one or more checkboxes to unselect.
         * Either Checkbox Control received these and discovered
         * it is not connected to the same file as Calc Helper
         * is or the Checkbox Control user has removed these boxes
         * from the 'Check Box Sets' list.
         * 
         * In this case we do not hassle with the filename as 
         * per the two cases above, one case the names differ,
         * another case they are the same but boxes are duplicates.
         * 
         * We pass a comma delimited string.
         */
        public void SendBoxUnSel(string boxUnSelStr)
        {
            if (client != null && client.Connected == true)
            {
                CTMessage msg = new CTMessage();
                msg.Command = "CBCBoxUnSel";
                msg.Data = boxUnSelStr;
                SendAsync(msg);
            }
        }

        /*
         * Tell Calc Helper no more check box selections.
         * The Calc Tool Server echos this command back because we
         * have a pending asychronous read hanging - we want to always
         * be ready to receive box selection messages when in that mode 
         * so we always have a ReceiveAsync() operating. This will
         * 'satisfy' that request, then we can 'ask again' for box
         * selections and be able to get the acknowledgement without
         * timing out.
         */
        public void BoxSelRequestOff()
        {
            if (client != null && client.Connected == true)
            {
                CTMessage msg = new CTMessage();
                msg.Command = "CBCBoxSelOff";

                SendAsync(msg);
            }
        }

        /*
         * Send Calc Helper the new Checkbox Control selected index.
         * We also use this function to tell Calc Helper the user
         * has just removed a checkbox set from setList.
         * 
         * The message text reads different from the way it is used
         * which is unfortunate, but required due to the inability
         * to edit the Calc Tool Server code.
         * 
         * Calc Helper maintains a copy of the setList and also the
         * selected index so that checkbox selections/unselections
         * appear the same for both tools.
         */
        public void SendSelectedIndexMsg(string msgData)
        {
            if (client != null && client.Connected == true)
            {
                CTMessage msg = new CTMessage();
                msg.Command = "CBCBoxSets";
                msg.FileName = vscFile;
                msg.Data = msgData;
                SendAsync(msg);
            }
        }

        // From Joe's SBClient:
        public void SendAsync(CTMessage msg)
        {
            if (client != null && client.Connected)
            {
                msg.Id = "CBC";
                string json = JsonConvert.SerializeObject(msg);

                byte[] bytes = Encoding.ASCII.GetBytes(json);

                byte[] head = BitConverter.GetBytes(bytes.Length);
                byte[] dmsg = new byte[bytes.Length + head.Length];
                head.CopyTo(dmsg, 0);
                bytes.CopyTo(dmsg, head.Length);

                // Begin sending the data to the remote device.  
                client.BeginSend(dmsg, 0, dmsg.Length, 0, new AsyncCallback(SendCallback), client);
            }
        }

        // From Joe's SBClient:
        public void SendCallback(IAsyncResult ar)
        {
            try
            {
                // Retrieve the socket from the state object.  
                Socket client = (Socket)ar.AsyncState;

                // Complete sending the data to the remote device.  
                int bytesSent = client.EndSend(ar);
                Console.WriteLine("Sent {0} bytes to server.", bytesSent);

                // Signal that all bytes have been sent.  
                sendMsgDone.Set();
            }
            catch (Exception e)
            {
                Console.WriteLine(e.ToString());
            }
        }

        public bool ReadPortNumber()
        {
            bool ret = false;

            try
            {
                string text = File.ReadAllText(portFile, Encoding.UTF8);
                ret = int.TryParse(text, out port);
            }
            catch (Exception ex)
            {
                ret = false;
                MessageBox.Show(ex.ToString(), "CTClient-ReadPortNumber():", MessageBoxButtons.OK, MessageBoxIcon.Error);
            }

            return ret;
        }

        public void ConnectCallback(IAsyncResult ar)
        {
            try
            {
                // Retrieve the socket from the state object.  
                Socket client = (Socket)ar.AsyncState;

                // Complete the connection.  
                client.EndConnect(ar);

                //MessageBox.Show(string.Format("Socket connected to {0}", client.RemoteEndPoint.ToString()), "Connected:", MessageBoxButtons.OK, MessageBoxIcon.Information);

                // Signal that the connection has been made.  
                connectDone.Set();
            }
            catch (Exception ex)
            {
                /*
                 * Pertinent comment from CalcHelper code:
                 * 
                 * Discovered tricky error condition while purposefully launching Calc Helper
                 * when the Calc Tool Server is not running. The form is blocked by the connectDone.WaitOne()
                 * call in StartClient(), so doing this and calling connectDone.Set() to kickstart execution
                 * but will be checking for connException to see if it was this situation.
                 */
                connException = true;

                MessageBox.Show(string.Format("{0}\n\n  This is required for commuication with\n  the Calc Helper tool.\n\n  Aborting Calc Tool Server Connection Attempt.", ex.ToString()), "CTClient-ConnectCallback():", MessageBoxButtons.OK, MessageBoxIcon.Error);

                connectDone.Set();
            }
            // Debug - Start:
            //finally
            //{
            //    //connException = true;
            //    //connectDone.Set();
            //}
            // Debug - Stop.
        }

        public void ReceiveAsync(Socket skt)
        {
            if (skt != null && skt.Connected)
            {
                InitStateObject isa = new InitStateObject();
                isa.workSocket = skt;

                try
                {
                    skt.BeginReceive(isa.LenBuffer, 0, isa.LenBuffer.Length, SocketFlags.None, new AsyncCallback(InitAsyncReadCallback), isa);
                }
                catch
                {

                }
            }
        }

        public void InitAsyncReadCallback(IAsyncResult ar)
        {
            try
            {
                InitStateObject isa = ar.AsyncState as InitStateObject;

                Socket client = isa.workSocket;

                int rcvd = client.EndReceive(ar);

                if (rcvd != 4)
                    throw new Exception();

                int to_recv = BitConverter.ToInt32(isa.LenBuffer, 0);

                // Create the state object.  
                StateObject state = new StateObject(to_recv);
                state.workSocket = client;
                state.toReceive = to_recv;

                client.BeginReceive(state.buffer, 0, state.buffer.Length, SocketFlags.None, new AsyncCallback(AsyncReadCallback), state);
            }
            catch
            {

            }
        }

        public void AsyncReadCallback(IAsyncResult ar)
        {
            StateObject state = ar.AsyncState as StateObject;
            Socket client = state.workSocket;

            int rcvd = client.EndReceive(ar);

            if (rcvd <= 0)
                return;

            state.BufStream.Write(state.buffer, 0, rcvd);

            state.toReceive -= rcvd;

            if (state.toReceive > 0)
            {
                Array.Clear(state.buffer, 0, state.buffer.Length);
                client.BeginReceive(state.buffer, 0, state.buffer.Length, SocketFlags.None, new AsyncCallback(AsyncReadCallback), state);
                return;
            }

            state.BufStream.Position = 0;

            string message = Encoding.ASCII.GetString(state.BufStream.ToArray());

            state.BufStream.Close();

            state.message = JsonConvert.DeserializeObject<CTMessage>(message);

            //MessageBox.Show(String.Format("Id: {0}, Command: {1}, FileName: {2}, Status: {3}", state.message.Id, state.message.Command, state.message.FileName, state.message.Status), "CTClient-AsyncReadCallback():", MessageBoxButtons.OK, MessageBoxIcon.Information);

            ProcessMessage(client, state.message);

            /* 
             * Commenting this out seems to work regarding sending proxy Check Box Control messages to
             * Cacl Helper proxy and getting the ProcessMessage() Ack msg...uncommented this seems to
             * somehow keep ProcessMessage() from not being called.
             */
            //receiveMsgDone.Reset();
            //ReceiveAsync(client);
        }

        public void ProcessMessage(Socket client, CTMessage message)
        {
            //string msgFile = "";

            CtClientArgs ctClientArgs;

            //MessageBox.Show("Entered ProcessMessage: " + message.Command, "CBC ProcessMessage():", MessageBoxButtons.OK, MessageBoxIcon.Information);
            //Debug.Print("Message: " + message.Command);
            //MessageBox.Show("Message: " + message.Command, "CTClient-ProcessMessage():", MessageBoxButtons.OK, MessageBoxIcon.Information);

            switch (message.Command)
            {
                case ("ConnectAck"):

                    if (message.Id.StartsWith("VSC") == true)
                    {
                        receiveMsgDone.Set();

                        //MessageBox.Show("Calc Tool Server has acknowledged the connection!", "CBC ProcessMessage():", MessageBoxButtons.OK, MessageBoxIcon.Information);

                        bConnected = true;

                        // We always have to have our ear on the wall listening...
                        ReceiveAsync(client);
                    }

                    break;

                case ("Disconnecting"):

                    if (message.Id.StartsWith("VSC") == true)
                    {
                        // Vscode is going down, bye bye time...
                        ctClientArgs = new CtClientArgs();
                        ctClientArgs.eventType = ctEvent.shutdown;
                        OnCtClientEvent(ctClientArgs);
                    }

                    break;

                /*
                 * Server echos this command back to CBC when CBC sends a
                 * "CBCBoxSelOff" command to CH. Awkward, but this keeps 
                 * us able to communicate thru the server to CalcHelper
                 * and back.
                 */
                case ("CBCBoxSelOff"):

                    if (message.Id.StartsWith("VSC") == true)
                    {
                        //MessageBox.Show("Server has echoed back the CBC 'CBCBoxSelOff' command.", "CTClient-ProcessMessage():", MessageBoxButtons.OK, MessageBoxIcon.Information);

                        ReceiveAsync(client);
                    }

                    break;

                case ("CBCBoxSelAck"):

                    if (message.Id.StartsWith("VSC") == true)
                    {
                        // Prevents a message timeout on receiving the Calc Helper Ack.
                        receiveMsgDone.Set();

                        boxSelFromCH = true;

                        if (checkboxAckWait == true)
                        {
                            checkboxAckWait = false;
                        }

                        //MessageBox.Show("CH has acknowledged the check box selection request!", "CTClient-ProcessMessage():", MessageBoxButtons.OK, MessageBoxIcon.Information);

                        /*
                         * Signal main form that Calc Helper has acknowledged our
                         * request for check box selections.
                         */
                        ctClientArgs = new CtClientArgs();
                        ctClientArgs.eventType = ctEvent.boxSelAck;
                        OnCtClientEvent(ctClientArgs);

                        // Setup for a check box selection message.
                        ReceiveAsync(client);
                    }

                    break;

                case ("BoxSelCBC"):

                    //MessageBox.Show(string.Format("  Box selections from Calc Helper:\n\n  message.FileName = {0}\n  message.Data = {1}\n  message.Id = {2}", message.FileName, message.Data, message.Id), "BoxSelCBC-ProcessMessage():", MessageBoxButtons.OK, MessageBoxIcon.Information);

                    if (message.Id.StartsWith("VSC") == true)
                    {
                        // As it turns out we are not checking the file names.
                        //msgFile = Path.GetFullPath(message.FileName).ToLower();

                        //if (msgFile == vscFile)
                        //{
                            // Tell the form about this.
                            ctClientArgs = new CtClientArgs();
                            ctClientArgs.eventType = ctEvent.boxSel;
                            ctClientArgs.eventMsg = message.Data;
                            OnCtClientEvent(ctClientArgs);
                        //}
                        //else
                        //{
                            // Let user know CH file name somehow differs, ignore boxes.
                            //MessageBox.Show(string.Format("  Check Box Control has received check boxes\n  from Calc Helper but they are from a different file name:\n\n  Calc Helper is connected to file: {0}\n\n  Check Box Control is connected to file: {1}\n\n  Boxes will be ignored.", msgFile, vscFile), "Calc Helper Check Box Selection Error:", MessageBoxButtons.OK, MessageBoxIcon.Error);
                            // Tell Calc Helper to unselect these boxes.
                            //SendBoxUnSel(message.Data);
                        //}

                        // Setup for another check box selection message.
                        ReceiveAsync(client);
                    }

                    break;

                case ("CBCBoxSelFileNack"):

                    if (message.Id.StartsWith("VSC") == true)
                    {
                        receiveMsgDone.Set();

                        MessageBox.Show(string.Format("  Calc Helper has rejected the request for\n  check box selections due to a difference in connected files:\n\n  Calc Helper is connected to file: {0}\n\n  Check Box Control is connected to file: {1}\n\n  Aborting request.", message.FileName, vscFile), "Check Box Control Box Selection Request Error:", MessageBoxButtons.OK, MessageBoxIcon.Error);

                        ReceiveAsync(client);
                    }

                    break;

                case ("CBCBoxSelNack"):

                    // This message comes from the Calc Tool Server.
                    if (message.Id.StartsWith("VSC") == true)
                    {
                        receiveMsgDone.Set();

                        MessageBox.Show("  Calc Helper is not connected to the Visual Studio Code\n  Calc Tool Server.\n\n  Start Calc Helper from Visual Studio Code and try again.", "Calc Helper Box Selection Request Error:", MessageBoxButtons.OK, MessageBoxIcon.Error);

                        ReceiveAsync(client);
                    }

                    break;

                case ("Disconnect-CH"):

                    //MessageBox.Show("  Calc Helper is closing.", "CH Disconnecting-ProcessMessage():", MessageBoxButtons.OK, MessageBoxIcon.Information);

                    // This message comes from the Calc Tool Server.
                    if (message.Id.StartsWith("VSC") == true)
                    {
                        if (checkboxAckWait == true)
                        {
                            // Probably never happen, but....
                            /*
                             * Prevents a message timeout on receiving the
                             * Calc Helper Ack from a box selection request.
                             */
                            receiveMsgDone.Set();

                            checkboxAckWait = false;
                        }

                        boxSelFromCH = false;

                        // Tell the form about this.
                        ctClientArgs = new CtClientArgs();
                        ctClientArgs.eventType = ctEvent.disconnectCH;

                        OnCtClientEvent(ctClientArgs);

                        ReceiveAsync(client);
                    }

                    break;

                case ("CHBoxSelOffCBC"):

                    //MessageBox.Show("  Calc Helper will stop sending checkbox selections.", "CH Box Selections OFF-ProcessMessage():", MessageBoxButtons.OK, MessageBoxIcon.Information);

                    if (message.Id.StartsWith("VSC") == true)
                    {
                        ctClientArgs = new CtClientArgs();

                        if (checkboxAckWait == true)
                        {
                            /*
                             * CalcHelper is denying the box selection request because
                             * either the document has no checkboxes on it or CalcHelper
                             * is in lockdown, i.e. not currently connected to the VSCode
                             * file it was launched from.
                             */

                            /*
                             * Prevents a message timeout on receiving the
                             * Calc Helper Ack from a box selection request.
                             */
                            receiveMsgDone.Set();

                            checkboxAckWait = false;

                            // Tell the form about this event:
                            ctClientArgs.eventType = ctEvent.docLockOrNoCheckboxes;
                        }
                        else
                        {
                            // Tell the form about this event:
                            ctClientArgs.eventType = ctEvent.boxSelOff;
                        }

                        boxSelFromCH = false;

                        OnCtClientEvent(ctClientArgs);

                        ReceiveAsync(client);
                    }

                    break;
            }
        }
    }

    public class VSCFileInfo
    {
        public string launchFileTmp = null;
        public string fileNameNoExtTmp = null;
        public string workFolderTmp = null;
        public string workFileTmp = null;
        public string fileTmp = null;
        public string fileNameTmp = null;
        public string portFileTmp = null;
        public bool portFileFound = false;
        public bool foundXmlFile = false;

        public verifyFile verifyResult;

        public VSCFileInfo()
        {
            portFileTmp = "C:\\temp\\CalcHelper\\.calctools";
        }

        /*
         * ChkBoxCtrl has been launched from Visual Studio with this file
         * argument - we must verify before we can continue with the launch
         * process.
         */
        public void VerifyFileInfo(string vscItemPath)
        {
            verifyResult = verifyFile.fileOk;

            try
            {
                if (string.IsNullOrEmpty(vscItemPath))
                {
                    verifyResult = verifyFile.nullEmptyPath;
                }
                else
                {
                    launchFileTmp = vscItemPath;
                    workFileTmp = Path.GetFileName(vscItemPath);

                    if (string.IsNullOrEmpty(workFileTmp))
                    {
                        verifyResult = verifyFile.nullEmptyFileName;
                    }
                    else 
                    {
                        if (Regex.Match(workFileTmp.ToLower(), @"^[a-z]{4}\d{4}\.js$").Success)
                        {
                            workFolderTmp = GetFolder(vscItemPath);

                            if (workFolderTmp != null)
                            {
                                /*
                                 * Verify existance of .calctool port number file:
                                 * 
                                 * portFileTmp has been initialized in VSCFileInfo's
                                 * constructor. Not a show stopper if not found, the
                                 * user would not be able to ask Calc Helper for box
                                 * selections in this case.
                                 */
                                if (File.Exists(portFileTmp))
                                {
                                    portFileFound = true;
                                }

                                /*
                                 * Retaining the exhaustive search where we first check for an
                                 * XML file either in the path of the code file being edited in
                                 * VSCode, else we check for an XML file in the current working
                                 * directory path. If no XML file was located we then do the
                                 * above again but now we would be looking for the document's ZIP
                                 * file.
                                 */
                                fileNameNoExtTmp = Path.GetFileNameWithoutExtension(workFileTmp).ToUpper();
                                fileNameTmp = fileNameNoExtTmp + ".xml";
                                fileTmp = Path.Combine(workFolderTmp, fileNameTmp);

                                if (File.Exists(fileTmp))
                                {
                                    foundXmlFile = true;
                                }
                                else
                                {
                                    fileTmp = Path.Combine(ChkBoxCtrl.Form1.curDir, fileNameTmp);

                                    if (File.Exists(fileTmp))
                                    {
                                        foundXmlFile = true;
                                    }
                                    else
                                    {
                                        // No XML file found, check for a ZIP file.
                                        fileNameTmp = fileNameNoExtTmp + ".zip";
                                        fileTmp = Path.Combine(workFolderTmp, fileNameTmp);

                                        if (!File.Exists(fileTmp))
                                        {
                                            // Actually expect it to be here 99.99999% of the time.
                                            fileTmp = Path.Combine(ChkBoxCtrl.Form1.curDir, fileNameTmp);

                                            //MessageBox.Show(string.Format("VerifyFileInfo - Found Zip: fileTemp = {0}", fileTmp));

                                            if (!File.Exists(fileTmp))
                                            {
                                                /*
                                                 * Hmmm, this may be a situation where the zip file is
                                                 * in a training folder of some kind/location, so caller
                                                 * will ask the user where the document's .zip file is.
                                                 */
                                                verifyResult = verifyFile.noZipFile;
                                            }
                                        }
                                    }
                                }
                            }
                            else
                            {
                                verifyResult = verifyFile.nullWorkFolder;
                            }
                        }
                        else
                        {
                            verifyResult = verifyFile.notDocFile;
                        }
                    }
                }
            }
            catch (Exception ex)
            {
                verifyResult = verifyFile.exception;
                MessageBox.Show(ex.ToString(), "CTClient-VerifyFileInfo():", MessageBoxButtons.OK, MessageBoxIcon.Error);
            }
        }

        public string GetFolder(string vscItemPath)
        {
            string folder = null;

            // Caller has ensured that vscEditorFile is not null.

            folder = Path.GetDirectoryName(vscItemPath);

            if (folder == null)
            {
                // Root path or the empty string.
                folder = Path.GetPathRoot(vscItemPath);
            }

            return folder;
        }
    }

    public class StateObject
    {
        public Socket workSocket = null;
        public byte[] buffer; // = new byte[BufferSize];

        public int toReceive;
        public MemoryStream BufStream;
        public CTMessage message;

        public StateObject(int toRec)
        {
            toReceive = toRec;
            BufStream = new MemoryStream(toRec);
            buffer = new byte[toRec];
        }
    }

    public class InitStateObject
    {
        public Socket workSocket = null;
        public byte[] LenBuffer = new byte[4];
    }

    public class CTMessage
    {
        public string Id { get; set; }
        public string Command { get; set; }
        public string FileName { get; set; }
        public string Status { get; set; }
        public string Data { get; set; }
        public Int32 hWnd;

        public CTMessage()
        {
            Id = "";
            Command = "";
            FileName = "";
            Status = "";
            Data = "";
            hWnd = 0;
        }
    }
}